#![warn(clippy::pedantic)]

pub mod is_prime;

use std::{time::SystemTime,thread, io::Write};
use num_bigint::BigUint;
use crate::is_prime::prime_utils;
use thread_priority::{set_current_thread_priority,ThreadPriority};

fn p(n: f64) -> f64 {
    let ln_n = n.ln();
    let ln_ln_n = ln_n.ln();
    let ln_ln_ln_n = ln_ln_n.ln();

    n * (
        ln_n + ln_ln_n - 1.0
        + (ln_ln_ln_n - 2.0) / ln_n
        - ((ln_ln_n).powi(2) - 6.0 * ln_ln_n + 11.0) / (2.0 * 2.0f64.log2() * ln_n)
        + ((ln_ln_n / ln_n).powi(3)) * (1.0 / ln_n)
    )
}


fn main() {

    if set_current_thread_priority(ThreadPriority::Max).is_err() {
        eprintln!("failed to set thread priority to max");
    }

    /* 
    let mut buffer = String::new();
    let stdin = io::stdin();
    print!("please enter the number to check: ");
    io::stdout().flush().expect("could not flush");
    stdin.read_line(&mut buffer).expect("could not read from stdin");
    buffer = buffer.trim().to_owned();
    
    let number = BigUint::parse_bytes(&buffer.as_bytes(), 10).expect("expected positive number");

    if is_prime(number.clone()) {
        println!("\n{number} is a prime");
    } else {
        println!("\n{number} is not a prime");
    }
    */

    let sys_time = SystemTime::now();

    const N:i64 = 100_000;
    const NU:usize = N as usize;
    let nf:f64 = 100_000f64;
    let nthprime_approx:usize = p(nf*1.02).ceil() as usize;
    const THREADS: usize = 12;
    let workloads = nthprime_approx/THREADS;

    let primecache: Vec<BigUint> = vec![];
    let primemutex = std::sync::Mutex::new(primecache);
    let primearc = std::sync::Arc::new(primemutex);
    let mut joins = vec![];
    for i in 0..THREADS {
        let arcclone = primearc.clone();
        let handle = thread::spawn(move || {

            if set_current_thread_priority(ThreadPriority::Max).is_err() {
                eprintln!("failed to set thread priority to max");
            }

            const SPLITTER: u128 = 10000u128;
            let one_thousand = &BigUint::from(SPLITTER*10);
            let one_hundred = &BigUint::from(SPLITTER);
            let ten = &BigUint::from(10u8);
            

            let work_uint = BigUint::from(workloads);
            let mut number = BigUint::from(i)*&work_uint;
            let inc = BigUint::from(1u8);
            // let zero = BigUint::from(0u8);
            
            let max = BigUint::from(i+1)*work_uint;
            let mut primecache = vec![];
            loop {
                number += &inc;
                
                if prime_utils::is_prime(&number,&primecache) {
                    primecache.push(number.clone());
                    //println!("Found a prime {} {}",number, primecache.len());
                }
                if (&number * one_thousand)/&max/ten == &number/&max*one_hundred {
                    let tried_getting = SystemTime::now();
                    let mut new_primecache = arcclone.lock().unwrap();
                    if SystemTime::now().duration_since(tried_getting).unwrap().as_millis() > 50 {
                        eprintln!("Waited 0.05+ seconds for lock!");
                    }
                    new_primecache.append(&mut primecache);
                    new_primecache.sort();
                    new_primecache.dedup();
                    
                    primecache = new_primecache.clone();
                    
                }
                if number == max {
                    break;
                }
            }   

            let mut new_primecache = arcclone.lock().unwrap();
            new_primecache.append(&mut primecache);
            new_primecache.sort();
            new_primecache.dedup();
        });
        joins.push(handle);
    }

    for handle in joins {
        handle.join().unwrap();
    }

    let mut primes = primearc.lock().unwrap().to_vec();

    let removed:i32 = if primes.len() > NU {
        let mut rem = 0;
        loop {
            if primes.len() <= NU {
                break;
            }
            rem+=1;
            primes.pop();
        }
        rem
    } else {
        0
    };

    let added = NU-primes.len();
    if primes.len() < NU {
        println!("Less primes than expected!");
        let mut number = BigUint::from(THREADS)*BigUint::from(workloads);
        let inc = BigUint::from(1u8);
        loop {
            number += &inc;
            if prime_utils::is_prime(&number, &primes) {
                primes.push(number.clone());
            }
            if primes.len() == NU {
                break;
            }
        }
    }

    let new_sys_time = SystemTime::now();
    let difference = new_sys_time.duration_since(sys_time)
    .expect("Clock may have gone backwards");

    assert_eq!(primes.len(),NU);

    println!("primes added: {} | % added: {}% | removed: {}",added,(added as f64)/primes.len() as f64*100.0,removed);

    println!("time: {difference:?}");

    println!("writing primes...");

    let mut file = std::fs::File::create("primes.json").unwrap();
    file.write_all(b"[").unwrap();
    let mut it = primes.iter().peekable();
    while let Some(prime) = it.next() {
        file.write_all(prime.to_string().as_bytes()).unwrap();
        if it.peek().is_some() {
            file.write_all(b",").unwrap();
        }
    }
    file.write_all(b"]").unwrap();



}
